function [x,y,p] = ReSNA(FUNC,nabFUNC,K,el,x0,y0,p0)

%%   Updated in September, 2013.  %%
%%   
%%   For more detailed usage, please see the manual_ReSNA.pdf.
%%  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%%    This program is based on the algorithm proposed in the
%%    following paper.
%%
%%    S. Hayashi, N. Yamashita and M. Fukushima, `A combined smoothing and
%%    regularization method for monotone second-order cone complementarity problems',
%%    SIAM Journal on Optimization, 15 (2005), pp. 593-615.
%%
%%    --- All program was coded by Shunsuke Hayashi. ---
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%  If you do not want to see the detaled progress, let PROGRESS = 'N'.

PROGRESS = 'Y';

%%   Values of parameters  %%

tole = 1e-8;      % This is used for termination criterion in Step 1.
tole_diff = 1e-8; % This is used to approximate the Jacobian by the finite difference method.
eta = 0.01;
eta_bar = 0.001;
rho = 0.5;
sigma = 0.4;
kappa = 0.01;
kappa_bar = 0.01;
kappa_hat = 1;


%%  Adjustment of absent K

n = sum(K);

if n==0
    K = [];
end
if isempty(K)==0
    if min(size(K)) > 1
        error('K must not be a matrix.');
    end
    if min(K)<1
        error('The values in K must not be less than 1.');
    end
    if sum(round(K)-K)~=0
        error('The values in K must be integers.');
    end
end


%% If [] is used for nabF, then it is approximated by the finite difference method.

if isempty(nabFUNC)
    nabFUNC = @(x)nabla_approx(FUNC,x,tole_diff,n+el,n+el);
    if PROGRESS == 'Y'
        fprintf('\n');
        fprintf('NOTICE: nabla F is approximated by the finite difference method.');
        fprintf('\n');
    end
end


%%  If [] is used for x0 or y0, then they are chosen randomly.

if nargin <= 6
    p0 = 2*rand(el,1) - 1;
    if nargin <= 5
        y0 = 2*rand(n,1) - 1;
        if nargin <=4
            x0 = 2*rand(n,1) - 1;
        end
    end
end
if isempty(x0)
    x0 = 2*rand(n,1) - 1;
end
if isempty(y0)
    y0 = 2*rand(n,1) - 1;
end
if isempty(p0)
    p0 = 2*rand(el,1) - 1;
end


%%  Dimension check

[x0raw,x0col] = size(x0);
[y0raw,y0col] = size(y0);
[p0raw,p0col] = size(p0);

if PROGRESS == 'Y'
    fprintf('\n');
    fprintf('============ (Input Information) =============\n');
    fprintf('Sum of input vector K (dimension of SOC) is %i.\n',sum(K));
    fprintf('The value of l (dimension of equality) is %i.\n',el);
    fprintf('Size of input vector x0 is (%i,%i).\n',x0raw,x0col);
    fprintf('Size of input vector y0 is (%i,%i).\n',y0raw,y0col);
    fprintf('Size of input vector p0 is (%i,%i).\n',p0raw,p0col);
    fprintf('Size of [x0;p0] is (%i,%i).\n',x0raw+p0raw,max(x0col,p0col));
end

[Fraw,Fcol] = size(FUNC([x0;p0]));
[nabFrow,nabFcol] = size(nabFUNC([x0;p0]));

if PROGRESS == 'Y'
    fprintf('Size of F is (%i,%i).\n',Fraw,Fcol);
    fprintf('Size of nabF is (%i,%i).\n',nabFrow,nabFcol);
    fprintf('==============================================\n');
end

dimdiff1 = sum(abs([x0raw,y0raw] - n));
dimdiff2 = sum(abs([Fcol,x0col,y0col,p0col] - 1));
dimdiff3 = sum(abs([x0raw+p0raw,Fraw,nabFrow,nabFcol] - (n+el)));

if dimdiff1+dimdiff2+dimdiff3 ~= 0
    fprintf('\n');
   error('Dimension Error!');
end

%%  Initialization in Step 0 %%

Hnrnm_now = norm(Hnr(FUNC,x0,y0,p0,K));
x = x0;
y = y0;
p = p0;
if isempty(K)
    mu = nan;
else
    mu = Hnrnm_now;
end
epsi = Hnrnm_now;
Hme_now = Hme(FUNC,x0,y0,p0,K,mu,epsi);
beta = norm(Hme_now);

%%  Other variables  %%

mu0_eta_bar = mu;
epsi0_eta_bar = epsi;
count_m = 0;
count_k = 0;
count_j = 0;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%   Main Algorithm    %%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if PROGRESS == 'Y'
        fprintf('\n');
        fprintf('\n');
    if isempty(K)
        fprintf('------------- (Iteration Information) --------------\n');
        fprintf('k j m mu    epsi      beta     |Hme|     |Hnr|\n');
    else
        fprintf('--------------- (Iteration Information) ----------------\n');
        fprintf('k j m     mu       epsi      beta     |Hme|     |Hnr|\n');
    end
end

%%%%%%%%%%    Step 1    %%%%%%%%%%

while Hnrnm_now > tole

    if PROGRESS == 'Y'
        fprintf('%u %u %u %3.3e %3.3e %3.3e %3.3e %3.3e\n',...
            count_k,count_j,count_m,mu,epsi,beta,...
            norm(Hme(FUNC,x,y,p,K,mu,epsi)),Hnrnm_now);
    end
    count_m = 0;
    flag_to_Step3 = 'N';

    %%%%%%%%%%   Step 2.1   %%%%%%%%%%

    direc = -Hmed(nabFUNC,x,y,p,K,mu,epsi)'\Hme_now;
    [dir_x, dir_y, dir_p] = makexyp(direc,n,el);
    xpdx = x + dir_x;
    ypdy = y + dir_y;
    ppdp = p + dir_p;
    Hmesq_now_dk = Hmesq(FUNC,xpdx,ypdy,ppdp,K,mu,epsi);

    %%%%%%%%%%   Step 2.2   %%%%%%%%%%

    if Hmesq_now_dk <= beta^2
        x = xpdx;
        y = ypdy;
        p = ppdp;
        flag_to_Step3 = 'Y';

        %%%%%%%%%%   Step 2.3   %%%%%%%%%%

    else
        rhomdx = dir_x;
        rhomdy = dir_y;
        rhomdp = dir_p;
        tw_sig_rhom = 2*sigma;
        Hmesq_vj_rhomdj = Hmesq_now_dk;
        Hmesq_vj = Hme_now'*Hme_now;
        while Hmesq_vj_rhomdj > (1-tw_sig_rhom)*Hmesq_vj
            tw_sig_rhom = tw_sig_rhom*rho;
            rhomdx = rhomdx*rho;
            rhomdy = rhomdy*rho;
            rhomdp = rhomdp*rho;
            Hmesq_vj_rhomdj = Hmesq(FUNC,x+rhomdx,y+rhomdy,p+rhomdp,K,mu,epsi);
            count_m = count_m + 1;
        end
        x = x + rhomdx;
        y = y + rhomdy;
        p = p + rhomdp;

        %%%%%%%%%%   Step 2.4   %%%%%%%%%%

        if Hmesq(FUNC,x,y,p,K,mu,epsi) <= beta^2
            flag_to_Step3 = 'Y';
        end
    end

    %  If the number of inner iterations exceeds 50, terminate.

    if count_j > 50
        Hnrnm_now = -Inf;
    else
        Hnrnm_now = norm(Hnr(FUNC,x,y,p,K));
    end

    %%%%%%%%%%    Step 3    %%%%%%%%%%

    if flag_to_Step3 == 'Y'
        count_j = 0;
        count_k = count_k + 1;
        mu0_eta_bar = eta_bar * mu0_eta_bar;
        epsi0_eta_bar = eta_bar * epsi0_eta_bar;
        
        if isempty(K)
            mu = nan;
        else
            mu = min([kappa*Hnrnm_now^2,mu0_eta_bar,...
                mu_bar(lam_til(x-y,K,Hnrnm_now*10^-4),kappa_hat*Hnrnm_now)]);
        end
        epsi = min(kappa_bar*Hnrnm_now^2, epsi0_eta_bar);
        beta = beta * eta;
    else
        count_j = count_j + 1;
    end
    Hme_now = Hme(FUNC,x,y,p,K,mu,epsi);
end

if PROGRESS == 'Y'
    fprintf('%u %u %u %3.3e %3.3e %3.3e %3.3e %3.3e\n',count_k,count_j,count_m,mu,epsi,beta,sqrt(Hmesq(FUNC,x,y,p,K,mu,epsi)),Hnrnm_now);
    if isempty(K)
        fprintf('----------------------------------------------------\n');
    else
        fprintf('--------------------------------------------------------\n');
    end
end
if Hnrnm_now == -Inf
    fprintf('\n');
    fprintf('CAUTION:\n');
    fprintf('Since the number of inner iterations exceeds 50,\n');
    fprintf('the iteration was forced to terminate.\n');
    fprintf('Obtained values are probably incorrect.\n');
end

%%%%%%%%%%% ALGORITHM END %%%%%%%%%%%%%



%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%      Definitions of additional functions 1     %%%%%
%%%%%%  -- functions for the sake of convenience --  %%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% 1 The following function partitions a 2n-dimensional vector into two n-dimensional vectors %%

function [x,y,p] = makexyp(w,n,l)
x = w(1:n);
y = w(n+1:2*n);
p = w(2*n+1:2*n+l);

%% 2 The following function partitions a vector corresponding to the Cartesian structure of K %%

function [array] = vec_part(z,K)
m = length(K);
dimend = 0;
array{m}=[];
for i=1:m
    dimst = dimend + 1;
    dimend = dimend + K(i);
    array{i} = z(dimst:dimend);
end

%% 3 The following function approximates the nabla F by means of the finite difference method.

function [val] = nabla_approx(FUNC,x,tole_diff,n,m)

val = zeros(m,n);
funcx = FUNC(x);

for ii=1:n
    ei = zeros(n,1);
    ei(ii) = 1;
    val(:,ii) = (FUNC(x + tole_diff*ei)-funcx)/tole_diff;
end

val = val';

%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%      Definitions of additional functions 2     %%%%%
%%%%%%   -- functions which appear in the paper --    %%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% 4  Square value of norm of H_{mu,epsi}  %%

function [val] = Hmesq(F,x,y,p,K,mu,epsi)
val = Hme(F,x,y,p,K,mu,epsi);
val = val'*val;

%% 5  Definition of H_{mu,epsi}  %%

function [val] = Hme(F,x,y,p,K,mu,epsi)

if isempty(K)
    val = F(p)+epsi*p;
else
    m = length(K);
    n = sum(K);
    l = length(p);
    x_part = vec_part(x,K);
    y_part = vec_part(y,K);
    
    val = nan(n,1);
    dimend = 0;
    
    for i=1:m
        dimst = dimend + 1;
        dimend = dimend + K(i);
        val(dimst:dimend) = phimu(x_part{i},y_part{i},mu);
    end
    val = [val; F([x;p])+epsi*[x;p]-[y;zeros(l,1)]];
end

%% 6  Definition of H_NR  %%

function [val] = Hnr(F,x,y,p,K)

if isempty(K) ~= 1
    
    m = length(K);
    n = sum(K);
    l = length(p);
    x_part = vec_part(x,K);
    y_part = vec_part(y,K);
    
    val = nan(n,1);
    dimend = 0;
    
    for i=1:m
        dimst = dimend + 1;
        dimend = dimend + K(i);
        val(dimst:dimend) = phinr(x_part{i},y_part{i});
    end
    val = [val; F([x;p])-[y;zeros(l,1)]];
    
else
    val = F(p);
end

%% 7  Definition of mu_bar  %%

function [val] = mu_bar(alpha,delta)
if delta>=0.5|alpha==0
    val = 10^10;
else
    val = abs(alpha)*sqrt(delta)/2;
end

%% 8  Definition of lambda_tilde  %%

function [val] = lam_til(z,K,tole)

m=length(K);
z_part = vec_part(z,K);
val = Inf;
for j=1:m
    zj = z_part{j};
    z1 = zj(1);
    zj(1) = 0;
    z2nm = norm(zj);
    lam_abs(1) = abs(z1-z2nm);
    lam_abs(2) = abs(z1+z2nm);
    for i=1:2
        if lam_abs(i) >= tole
            val = min(val,lam_abs(i));
        end
    end
end
if val == Inf
    val = 0;
end

%% 9  Definition of phi_mu  %%

function [val] = phimu(x,y,mu)
z = x - y;
z1 = z(1);
z2nm = sqrt(z'*z - z1^2);
sqmns = sqrt((z1-z2nm)^2+4*mu^2);
sqpls = sqrt((z1+z2nm)^2+4*mu^2);
val = x-(0.5+z1/(sqpls+sqmns))*z;
val(1) = x(1)-(2*z1+sqmns+sqpls)/4;

%% 10  Definition of phi_NR  %%

function [val] = phinr(x,y)
z = x - y;
z1 = z(1);
z2nm = sqrt(z'*z - z1^2);
if z1 >= z2nm
    val = y;
elseif -z1 >= z2nm
    val = x;
else
    val = x - 0.5*(1+z1/z2nm)*z;
    val(1) = x(1) - 0.5*(z1+z2nm);
end


%% 11   Definition of nabla H_{mu,epsi}  %%

function [val] = Hmed(nabF,x,y,p,K,mu,epsi)

if isempty(K)
    val = nabF(p) + epsi*eye(length(p));
else
    m = length(K);
    z = x-y;
    n = length(z);
    l = length(p);
    z_part = vec_part(z,K);
    Pmudxy = zeros(n);
    dimend = 0;
    
    for i=1:m
        dimst = dimend + 1;
        dimend = dimend + K(i);
        Pmudxy(dimst:dimend,dimst:dimend) = Pmud(z_part{i},mu);
    end
    
    nabFval_ep = nabF([x;p]) + epsi*eye(n+l);
    
    val = [eye(n)-Pmudxy, nabFval_ep(1:n,:);
        Pmudxy, [-eye(n), zeros(n,l)];
        zeros(l,n), nabFval_ep(n+1:n+l,:)];
end

%% 12  Definition of nabla P_mu  %%

function [val] = Pmud(z,mu)
len = length(z);
z1 = z(1);
z2 = zeros(len-1,1);
for k=2:len
    z2(k-1) = z(k);
end;
z2nm = norm(z2);
if z2nm==0
    val = gamud(z1,mu)*eye(len);
else
    uniz2 = z2/z2nm;
    lmd1 = z1 - z2nm;
    lmd2 = z1 + z2nm;
    gamudlmd1 = gamud(lmd1,mu);
    gamudlmd2 = gamud(lmd2,mu);
    a = amu(lmd1,lmd2,mu);
    b = 0.5*(gamudlmd2 + gamudlmd1);
    c = 0.5*(gamudlmd2 - gamudlmd1);
    val = [b c*uniz2';c*uniz2 a*eye(len-1)+(b-a)*uniz2*uniz2'];
end

%% 13  Definition of derivative of gamma_mu  %%

function [val] = gamud(lmd,mu)
val = 0.5*(1 + lmd/sqrt(lmd^2+4*mu^2));

%% 14  Definition of a_mu in (3.15)  %%
%%  -- multiplied the numerator and the denominator
%%     by sqrt(lmd1^2+4*mu^2)+sqrt(lmd2^2+4*mu^2) --

function [val] = amu(lmd1,lmd2,mu)
val = 0.5 * (1 + (lmd1+lmd2)/(sqrt(lmd1^2+4*mu^2)+sqrt(lmd2^2+4*mu^2)));
